/*
 * Copyright (C) 2013 Christopher Burke <krslynx@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package org.xproduct.server.core.net;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.Queue;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.vertx.java.core.http.ServerWebSocket;
import org.xproduct.server.core.Game;
import org.xproduct.server.core.exceptions.GameLogicException;
import org.xproduct.server.core.exceptions.GameProtocolException;
import org.xproduct.server.core.model.Player;

/**
 *
 * @author Christopher Burke <krslynx@gmail.com>
 */
public abstract class PacketController {

    /**
     * The logger for the class
     */
    private Logger log = LoggerFactory.getLogger(this.getClass().getSuperclass().getName());
    /**
     * The incoming object array, which is an immediate translation of the JSON
     * into Java
     */
    public Object[] incomingObject;
    /**
     * The gameServer, which contains the player data
     */
    public Game gameServer;
    /**
     * The WebSocket
     */
    private final ServerWebSocket webSocket;

    /**
     * The PacketController constructor, takes in as arguments, the incoming
     * object, game server and the socket identifier. This allows for data to be
     * processed through implementation classes
     *
     * @param incomingObject the incoming object, a direct conversion of the
     * JSON data
     * @param gameServer The Game Server which contains the player data etc
     * @param socketId The Socket Identifier for the Player
     */
    public PacketController(Object[] incomingObject, Game gameServer, ServerWebSocket webSocket) {
        this.incomingObject = incomingObject;
        this.gameServer = gameServer;
        this.webSocket = webSocket;
    }

    /**
     * Called to process the incoming object, game server and socket id
     *
     * @throws GameProtocolException if there is an error processing
     */
    public abstract void process() throws GameProtocolException;

    /**
     * Send data to the WebSocket as a response (Will only go to the server)
     *
     * @param data the data to send to the socket
     */
    public void send(String data) {
        webSocket.writeTextFrame(data);
    }

    /**
     * Send data to all WebSockets apart from this one
     * @param data the data to send to the sockets
     */
    public void sendToAll(String data) {
        HashMap<ServerWebSocket, Player> playerMap = gameServer.getPlayerMap();
        Iterator<Entry<ServerWebSocket, Player>> playerMapIterator = playerMap.entrySet().iterator();
        while (playerMapIterator.hasNext()) {
            Entry<ServerWebSocket, Player> entry = playerMapIterator.next();
            if (entry.getKey() != webSocket) {
                entry.getKey().writeTextFrame(data);
            }
        }
    }
    
    /**
     * Get the player for the session
     * @return the Player
     */
    public Player getPlayer() {
        HashMap<ServerWebSocket, Player> playerMap = gameServer.getPlayerMap();
        if(playerMap.containsKey(webSocket)) {
            return playerMap.get(webSocket);
        } else {
            log.error("Failed to find player by socket identifier: " + webSocket.textHandlerID);
            return null;
        }
    }
}
