/*
 * Copyright (C) 2013 Christopher Burke <krslynx@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package org.xproduct.server.core.net.controllers;

import com.google.gson.Gson;
import com.google.gson.JsonSyntaxException;
import com.google.gson.internal.StringMap;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.Map;
import org.json.simple.parser.ParseException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.vertx.java.core.http.ServerWebSocket;
import org.xproduct.server.core.Game;
import org.xproduct.server.core.exceptions.GameProtocolException;
import org.xproduct.server.core.model.Player;
import org.xproduct.server.core.model.social.FacebookIdentifier;
import org.xproduct.server.core.model.team.Team;
import org.xproduct.server.core.net.PacketController;
import org.xproduct.server.core.net.packets.S2CLoginResponse;
import org.xproduct.server.core.net.packets.S2CPacketType;

/**
 *
 * @author Christopher Burke <krslynx@gmail.com>
 */
public class LoginController extends PacketController {

    /**
     * The logger for the class
     */
    private Logger log = LoggerFactory.getLogger(LoginController.class);
    private String accessToken;
    private boolean tokenValid = true;
    private boolean offlinePlayer = false;

    public LoginController(Object[] incomingObject, Game gameServer, ServerWebSocket webSocket) {
        super(incomingObject, gameServer, webSocket);
    }

    /**
     * {@inheritDoc}
     */
    public void process() throws GameProtocolException {


        parseAccessToken(incomingObject);

        String id;
        String name;

        if (!offlinePlayer) {
            try {
                String[] facebookCredentials = getFacebookCredentials(accessToken);
                id = facebookCredentials[0];
                name = facebookCredentials[1];
            } catch (MalformedURLException ex) {
                tokenValid = false;
                throw new GameProtocolException("Failure to process access token.", ex);
            } catch (IOException ex) {
                tokenValid = false;
                throw new GameProtocolException("Failure to process access token.", ex);
            }
        } else {
            id = "1231123112";
            name = "OFFLINE PLAYER";
        }


        Player playerForSocket;

        playerForSocket = getPlayer();
        if (playerForSocket == null) {
            return;
        }

        if (!tokenValid) {
            try {
                Map gameSettings = gameServer.getGameSettings();

                Object[] returnObj = new Object[]{
                    S2CPacketType.LOGIN_REPLY.getPacketTypeIdentifier(),
                    S2CLoginResponse.LOGIN_FAILURE.getPacketTypeIdentifier(),
                    gameServer.getGameResources(),
                    gameSettings,
                    gameServer.getGameMapData(),
                    gameServer.getGameMapProperties()
                };
                send(new Gson().toJson(returnObj));
            } catch (ParseException ex) {
                throw new GameProtocolException("Exception when attempting to parse data.", ex);
            }
        } else {
            try {
                Map gameSettings = gameServer.getGameSettings();
                gameSettings.put("id", id);
                gameSettings.put("name", name);
                gameSettings.put("team", 121);
                gameSettings.put("showTutorial", true);

                playerForSocket.setSocialIdentifier(new FacebookIdentifier(id, name));
                playerForSocket.setTeam(new Team((short) 1231));
                playerForSocket.setShowTutorial(true);


                Object[] returnObj = new Object[]{
                    S2CPacketType.LOGIN_REPLY.getPacketTypeIdentifier(),
                    S2CLoginResponse.LOGIN_SUCCESS.getPacketTypeIdentifier(),
                    gameServer.getGameResources(),
                    gameSettings,
                    gameServer.getGameMapData(),
                    gameServer.getGameMapProperties()
                };
                send(new Gson().toJson(returnObj));
            } catch (ParseException ex) {
                throw new GameProtocolException("Exception when attempting to parse data.", ex);
            }
        }

    }

    private String parseAccessToken(Object[] incomingObject) throws GameProtocolException {

        log.debug("Attempting to retrieve access token from login packet.");

        String stringFromObject = null;

        try {
            StringMap sm = (StringMap) incomingObject[1];
            this.accessToken = (String) sm.get("accessToken");
        } catch (ArrayIndexOutOfBoundsException ex) {
            tokenValid = false;
            throw new GameProtocolException("There appears to be no login information inside the login packet.", ex);
        } catch (JsonSyntaxException ex) {
            tokenValid = false;
            throw new GameProtocolException("Failure when attempting retrieve login data from JSON.", ex);
        }

        log.debug("Access token retrieved: " + this.accessToken);

        if (this.accessToken == null) {
            if (gameServer.isDevelopment()) {
                log.debug("The access token received was null, however server is in development mode. Processing as an offline request.");
                offlinePlayer = true;
            } else {
                log.error("Failure to retreieve the access token for the player. Refusing connection.");
                throw new GameProtocolException("Connection refused due to malformed login credentials.");
            }
        }

        return stringFromObject;
    }

    private String[] getFacebookCredentials(String accessToken) throws MalformedURLException, IOException {
        // TODO: Will probably change this to HttpClient or something...
        
        URL facebookMe = new URL("https://graph.facebook.com/me?access_token=" + accessToken);
        URLConnection urlConnection = facebookMe.openConnection();
        BufferedReader bufferedReader = new BufferedReader(
                new InputStreamReader(
                urlConnection.getInputStream()));
        
        String inputLine;
        String response = "";
        while ((inputLine = bufferedReader.readLine()) != null) {
            response += inputLine;
    }
        bufferedReader.close();

        FacebookResponse facebookResponse = new Gson().fromJson(response, FacebookResponse.class);

        return new String[]{facebookResponse.id, facebookResponse.name};
    }

    class FacebookResponse {

        String id;
        String name;
    }
}
