/*
 * Copyright (C) 2013 Christopher Burke <krslynx@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package org.xproduct.server.vertx;

import com.google.gson.Gson;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.Timer;
import java.util.TimerTask;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.vertx.java.core.Handler;
import org.vertx.java.core.buffer.Buffer;
import org.vertx.java.core.http.HttpServerRequest;
import org.vertx.java.core.http.ServerWebSocket;
import org.vertx.java.deploy.Verticle;
import org.xproduct.server.core.Game;
import org.xproduct.server.core.exceptions.GameProtocolException;
import org.xproduct.server.core.model.Player;
import org.xproduct.server.core.net.packets.core.s2c.S2CPlayerLeft;

/**
 *
 * @author Christopher Burke <krslynx@gmail.com>
 */
public class Server extends Verticle {

    /**
     * The path for WebSockets
     */
    public static String websocketPath = "/dotproduct/v1/";
    /**
     * The relative path to the HTTP (game client) directory
     */
    public static String httpFilePath = "src/main/resources/dotproduct-client/";
    /**
     * The relative path from the httpFilePath to the resources.json file
     */
    public static String resourcesPath = "/data/arenas/svs/resources.json";
    /**
     * The relative path from the httpFilePath to the settings.json file
     */
    public static String settingsPath = "/data/arenas/svs/settings.json";
    /**
     * The relative path from the httpFilePath to the map.json file
     */
    public static String mapPath = "/data/arenas/svs/map.json";
    /**
     * The relative path from the httpFilePath to the tile_properties.json file
     */
    public static String mapPopertiesPath = "/data/arenas/svs/tile_properties.json";
    /**
     * The port the server should run on
     */
    public static int serverPort = 8000;
    /**
     * The host the server should run on
     */
    public static String serverHost = "localhost";
    /**
     * If the server is in development mode
     */
    public static boolean development = true;
    /**
     * The logger for the class
     */
    private Logger log = LoggerFactory.getLogger(Server.class);
    /**
     * Variable space for the compiled client JavaScript code
     */
    private static String applicationJs;
    /**
     * Variable space for the gameResources JSON data
     */
    private static String gameResources;
    /**
     * Variable space for the gameSettings JSON data
     */
    private static String gameSettings;
    /**
     * variable space for the gameMapData JSON data
     */
    private static String gameMapData;
    /**
     * Variable space for the gameMapProperties JSON data
     */
    private static String gameMapProperties;
    /**
     * Variable space for the Core Game server (handles sockets etc)
     */
    private Game gameServer;

    /**
     * Called when the Verticle is initiated, and constructs the HTTP,
     * WebSockets and sets up the game settings data.
     *
     * @throws Exception if there is a Vertx.io issue
     */
    @Override
    public void start() throws Exception {

        log.info("Reading server settings...");
        setupResources("serverSettings.json");
        log.info("Reading server settings completed.");

        log.info("Starting " + Server.class.getName() + "...");

        try {
            if (gameResources == null) {
                log.debug("Loading game resources.");
                gameResources = readFile(httpFilePath + resourcesPath);
                log.debug("Game resources loaded.");
            }

            if (gameSettings == null) {
                log.debug("Loading game settings.");
                gameSettings = readFile(httpFilePath + settingsPath);
                log.debug("Game settings loaded.");
            }

            if (gameMapData == null) {
                log.debug("Loading game map data.");
                gameMapData = readFile(httpFilePath + mapPath);
                log.debug("Game map data loaded.");
            }

            if (gameMapProperties == null) {
                log.debug("Loading game map data.");
                gameMapProperties = readFile(httpFilePath + mapPopertiesPath);
                log.debug("Game map data loaded.");
            }

            if (applicationJs == null) {
                log.debug("Compiling the dotproduct JavaScript code using the Closure compiler.");

                String applicationJsFile = "applicationJs.js.temp";
                if (development && new File(applicationJsFile).exists()) {
                    log.debug("In development mode, attempting to read the stored compiled js code.");
                    applicationJs = readFile(applicationJsFile, true);
                    log.debug("Stored compiled js code read.");
                } else {
                    applicationJs = compileClient();
                    if (development) {
                        log.debug("In development mode, attempting to write the stored compiled js code.");
                        writeFile(applicationJsFile, applicationJs);
                        log.debug("Stored compiled js code.");
                    }
                }
                log.debug("JavaScript code compiled.");
            }

        } catch (IOException ex) {
            log.error("An IOException occurred when attempting to process game files. Cannot continue.", ex);
            System.exit(-1);
        }

        log.debug("Constructing " + Game.class.getName());
        gameServer = new Game(gameResources, gameSettings, gameMapData, gameMapProperties);
        if (development) {
            log.info("Setting environment to Development.");
            gameServer.setToDevelopment();
        } else {
            log.info("Setting environment to Production.");
        }
        log.debug(Game.class.getName() + " constructed.");

        log.debug("Starting Vertx.io HTTP and WebSocket servers....");
        vertx.createHttpServer()
                .setAcceptBacklog(10000)
                .websocketHandler(getServerWebSocketHandler())
                .requestHandler(getHttpServerRequestHandler())
                .listen(serverPort, serverHost);
        log.info("HTTP and WebSocket servers started; listening for connections on port: " + serverPort);

    }

    /**
     * @return the handler for HTTP requests
     */
    private Handler<HttpServerRequest> getHttpServerRequestHandler() {
        return new Handler<HttpServerRequest>() {
            public void handle(HttpServerRequest req) {
                log.debug("Request for HTTP content at path: " + req.path);
                if (req.path.equals("/js/Application.js")) {
                    req.response.putHeader("Content-Length", applicationJs.length());
                    req.response.putHeader("Content-Type", "text/javascript; charset=utf-8");
                    req.response.write(applicationJs);
                    req.response.end();
                } else {
                    String path = req.path.equals("/") ? "index.html" : req.path;
                    req.response.sendFile(httpFilePath + path);
                }
                log.debug("Http content served.");
            }
        };
    }

    /**
     * @return the handler for WebSocket requests
     */
    private Handler<ServerWebSocket> getServerWebSocketHandler() {
        return new Handler<ServerWebSocket>() {
            public void handle(ServerWebSocket ws) {
                log.debug("Request for WebSocket " + ws.textHandlerID + " at path: " + ws.path);
                if (ws.path.equals(websocketPath)) {
                    log.debug("Adding WebSocket Connection " + ws.textHandlerID + " to connection list.");
                    addConnection(ws);
                    log.debug("WebSocket connection " + ws.textHandlerID + " added.");
                } else {
                    log.debug("Rejecting WebSocket: " + ws.textHandlerID + " from connecting.");
                    ws.reject();
                    log.debug("Rejected WebSocket: " + ws.textHandlerID);
                }
            }
        };
    }

    private void addConnection(final ServerWebSocket webSocket) {
        Player player = gameServer.addPlayer(webSocket);
        webSocket.dataHandler(new Handler<Buffer>() {
            public void handle(Buffer e) {
                gameServer.processRequest(e.toString(), webSocket);
            }
        });
        webSocket.endHandler(new Handler<Void>() {
            public void handle(Void e) {
                
                HashMap<ServerWebSocket, Player> playerMap
                        = gameServer.getPlayerMap();
                
                Player leavingPlayer = playerMap.get(webSocket);
                
                Iterator<Entry<ServerWebSocket, Player>> playerMapIterator 
                        = playerMap.entrySet().iterator();
                
                while(playerMapIterator.hasNext()) {
                    Entry<ServerWebSocket, Player> entry = playerMapIterator.next();
                    ServerWebSocket socket = entry.getKey();
                    if(socket != webSocket) {
                        // yeah... nasty...
                        socket.writeTextFrame(new S2CPlayerLeft(gameServer.getPlayerBySocialIdentifier(leavingPlayer.getSocialIdentifier().getId()).getValue().getSocialIdentifier().getId()).getPacket());
                    }
                }
            }
        });
    }

    /**
     * Compiles the dotproduct client code, and returns the compiled code
     * generated by the closure compiler
     *
     * @return the compiled closure compiler
     * @throws IOException if there was an error reading files
     */
    private String compileClient() throws IOException {

        String outStream = null;
        String errStream = null;
        String compilerOutput = "";
        String errorOutput = "";

        Process p = Runtime.getRuntime().exec(new String[]{
                    "python",
                    httpFilePath + "closure/bin/calcdeps.py",
                    "-i",
                    httpFilePath + "js/Application.js",
                    "-o",
                    "script",
                    "--compiler_jar",
                    httpFilePath + "../compiler.jar",
                    "-p",
                    httpFilePath + "closure", "-p",
                    httpFilePath + "js"
                });

        BufferedReader stdIn = new BufferedReader(new InputStreamReader(p.getInputStream()));

        BufferedReader stdErr = new BufferedReader(new InputStreamReader(p.getErrorStream()));

        while ((outStream = stdIn.readLine()) != null) {
            if (outStream != null) {
                compilerOutput += outStream + "\n";
            }

        }

        while ((errStream = stdErr.readLine()) != null) {
            if (errStream != null) {
                errorOutput += errStream + "\n";
            }
        }

        if (!errorOutput.isEmpty()) {
            log.debug("Error stream was not empty following compilation of dotproduct client, error stream data: " + errorOutput);
        }

        return compilerOutput;
    }

    private String readFile(String file) throws IOException {
        return readFile(file, false);
    }

    /**
     * Reads a file, and returns the content
     *
     * @param file the file
     * @return the content
     * @throws IOException if there was an issue reading the file
     */
    private String readFile(String file, boolean lineBreaks) throws IOException {

        BufferedReader fileIn = null;
        StringBuilder fileOutput = new StringBuilder();

        log.debug("Reading file: " + file);

        try {
            fileIn = new BufferedReader(new FileReader(file));

            String sCurrentLine;
            while ((sCurrentLine = fileIn.readLine()) != null) {
                fileOutput.append(sCurrentLine);
                if (lineBreaks) {
                    fileOutput.append("\n");
                }
            }

        } finally {
            if (fileIn != null) {
                fileIn.close();
            }
        }

        log.debug("File read: " + file);

        return fileOutput.toString();
    }

    /**
     * Writes a file, with provided content
     *
     * @param fileUri the file to write
     * @param content the content to write into it
     * @throws IOException if there was an issue writing to the file
     */
    private void writeFile(String fileUri, String content) throws IOException {

        File file = new File(fileUri);

        log.debug("Writing file: " + fileUri);

        if (!file.exists()) {
            file.createNewFile();
            log.debug("File created: " + fileUri);
        }

        FileWriter fw = new FileWriter(file.getAbsoluteFile());
        BufferedWriter bw = new BufferedWriter(fw);
        bw.write(content);
        bw.close();

        log.debug("File written: " + fileUri);
    }

    /**
     * Sets up the resources, from the file
     *
     * @param fileUri the file to read data from
     * @throws IOException if there was an exception reading the file
     */
    private void setupResources(String fileUri) throws IOException {

        File file = new File(fileUri);

        log.info("Checking if resources file " + fileUri + " exists.");

        if (!file.exists()) {
            log.info("Resources file did not exist. Sadface. Creating one.");

            Resources resources = new Resources();
            resources.websocketPath = websocketPath;
            resources.httpFilePath = httpFilePath;
            resources.resourcesPath = resourcesPath;
            resources.settingsPath = settingsPath;
            resources.mapPath = mapPath;
            resources.mapPopertiesPath = mapPopertiesPath;
            resources.serverPort = serverPort;
            resources.serverHost = serverHost;

            String resourcesJson = new Gson().toJson(resources);
            writeFile(fileUri, resourcesJson);
            log.info("Resources have been written to file.");
        } else {
            log.info("Resources file exists.");
            String resourcesJson = readFile(fileUri);
            Resources resources = new Gson().fromJson(resourcesJson, Resources.class);
            log.info("Resources JSON file converted to Java.");

            if (resources.websocketPath != null) {
                websocketPath = resources.websocketPath;
                log.debug("Set WebSocket path: " + websocketPath);
            }

            if (resources.httpFilePath != null) {
                httpFilePath = resources.httpFilePath;
                log.debug("Set Http File path: " + httpFilePath);
            }

            if (resources.resourcesPath != null) {
                resourcesPath = resources.resourcesPath;
                log.debug("Set Resources path: " + resourcesPath);
            }

            if (resources.settingsPath != null) {
                settingsPath = resources.settingsPath;
                log.debug("Set Settings path: " + settingsPath);
            }

            if (resources.mapPath != null) {
                mapPath = resources.mapPath;
                log.debug("Set Map path: " + mapPath);
            }

            if (resources.mapPopertiesPath != null) {
                mapPopertiesPath = resources.mapPopertiesPath;
                log.debug("Set Map Properties path: " + mapPopertiesPath);
            }

            if (resources.serverPort != 8000) {
                serverPort = resources.serverPort;
                log.debug("Set Server Port: " + serverPort);
            }

            if (resources.serverHost != null) {
                serverHost = resources.serverHost;
                log.debug("Set Server Host: " + serverHost);
            }

            if (resources.development) {
                development = resources.development;
                log.debug("Set Server Development mode to: " + development);
            }
        }

    }
}
